from .commandenv import ResultRow
from .exceptions import CommandLineError, NoDataError
from .parsing import (
    BlackMarketSwitch, FleetCarrierArgument, MutuallyExclusiveGroup,
    NoPlanetSwitch, OdysseyArgument, PadSizeArgument, ParseArgument,
    PlanetaryArgument,
)
from itertools import chain
from ..tradedb import TradeDB, System, Station, describeAge
from ..tradecalc import TradeCalc, Route, NoHopsError

import math
import sys
import time


######################################################################
# Parser config

help = 'Calculate best trade run.'
name = 'run'
epilog = None
usesTradeData = True

arguments = [
    ParseArgument('--capacity',
            help = 'Maximum capacity of cargo hold.',
            metavar = 'N',
            type = int,
        ),
    ParseArgument('--credits',
            help = 'Starting credits.',
            metavar = 'CR',
            type = "credits",
        ),
]

switches = [
    ParseArgument('--from', '-f',
        help = 'Starting system/station.',
        dest = 'starting',
        metavar = 'STATION',
    ),
    MutuallyExclusiveGroup(
        ParseArgument('--to', '-t',
            help = 'Final system/station.',
            dest = 'ending',
            metavar = 'PLACE',
            default = None,
        ),
        ParseArgument('--towards', '-T',
            help = (
                'Choose a route that continually reduces the '
                'distance towards this system.'
            ),
            dest = 'goalSystem',
            metavar = 'SYSTEM',
            default = None,
        ),
        ParseArgument('--loop',
            help = 'Return to the starting station.',
            action = 'store_true',
            default = False,
        ),
    ),
    ParseArgument('--via',
        help = 'Require specified systems/stations to be en-route.',
        action = 'append',
        metavar = 'PLACE[,PLACE,...]',
    ),
    ParseArgument('--avoid',
        help = 'Exclude an item, system or station from trading. '
                'Partial matches allowed, '
                'e.g. "dom.App" or "domap" matches "Dom. Appliances".',
        action = 'append',
    ),
    MutuallyExclusiveGroup(
        ParseArgument('--direct',
            help = "Assume destinations are reachable without worrying "
                "about jumps.",
            action = 'store_true',
        ),
        ParseArgument('--hops',
            help = 'Number of hops (station-to-station) to run.',
            default = 2,
            type = int,
            metavar = 'N',
        ),
    ),
    ParseArgument('--jumps-per',
        help = 'Maximum number of jumps (system-to-system) per hop.',
        default = 2,
        dest = 'maxJumpsPer',
        metavar = 'N',
        type = int,
    ),
    ParseArgument('--ly-per',
        help = 'Maximum light years per jump.',
        dest = 'maxLyPer',
        metavar = 'N.NN',
        type = float,
    ),
    ParseArgument('--empty-ly',
        help = 'Maximum light years ship can jump when empty.',
        dest = 'emptyLyPer',
        metavar = 'N.NN',
        type = float,
        default = None,
    ),
    ParseArgument('--start-jumps', '-s',
        help = 'Consider stations within this many jumps of the origin '
             '(requires --from).',
        dest = 'startJumps',
        default = 0,
        type = int,
    ),
    ParseArgument('--end-jumps', '-e',
        help = 'Consider stations within this many jumps of the destination '
             '(requires --to).',
        dest = 'endJumps',
        default = 0,
        type = int,
    ),
    ParseArgument('--show-jumps', '-J',
        help = 'Show detail of jumps between hops.',
        dest = 'showJumps',
        action = 'store_true',
    ),
    ParseArgument('--limit',
        help = 'Maximum units of any one cargo item to buy (0: unlimited).',
        metavar = 'N',
        type = int,
    ),
    ParseArgument('--age', '--max-days-old', '-MD',
        help = 'Maximum age (in days) of trade data to use.',
        metavar = 'DAYS',
        type = float,
        dest = 'maxAge',
    ),
    PadSizeArgument(),
    MutuallyExclusiveGroup(
        NoPlanetSwitch(),
        PlanetaryArgument(),
    ),
    FleetCarrierArgument(),
    OdysseyArgument(),
    BlackMarketSwitch(),
    ParseArgument('--ls-penalty', '--lsp',
        help = "Penalty per 1kls stations are from their stars.",
        default = 12.5,
        type = float,
        dest = 'lsPenalty'
    ),
    ParseArgument('--ls-max',
        help = 'Only consider stations upto this many ls from their star.',
        metavar = 'LS',
        dest = 'maxLs',
        type = int,
        default = 0,
    ),
    ParseArgument('--gain-per-ton', '--gpt',
        help = 'Specify the minimum gain per ton of cargo',
        dest = 'minGainPerTon',
        type = "credits",
        default = 1
    ),
    ParseArgument('--max-gain-per-ton', '--mgpt',
        help = 'Specify the maximum gain per ton of cargo',
        dest = 'maxGainPerTon',
        type = "credits",
        default = 0
    ),
    ParseArgument('--unique',
        help = 'Only visit each station once.',
        action = 'store_true',
        default = False,
    ),
    ParseArgument('--loop-interval', '-li',
        help = (
            'Require this many hops between visits to the same station. '
            'A value of 1 would be the default behavior, so a value of '
            '2 is the minimum allowed.'
        ),
        type = int,
        default = None,
        dest = 'loopInt',
    ),
    ParseArgument('--margin',
        help = 'Reduce gains made on each hop to provide a margin of error '
                'for market fluctuations (e.g: 0.25 reduces gains by 1/4). '
                '0<: N<: 0.25.',
        default = 0.00,
        metavar = 'N.NN',
        type = float,
    ),
    ParseArgument('--insurance',
        help = 'Reserve at least this many credits to cover insurance.',
        default = 0,
        metavar = 'CR',
        type = "credits",
    ),
    ParseArgument('--routes',
        help = 'Maximum number of routes to show. DEFAULT: 1',
        default = 1,
        metavar = 'N',
        type = int,
    ),
    ParseArgument('--max-routes',
        help = 'At the end of each hop, limit the number of routes '
                'that continue to the next round to the top N '
                'highest scoring',
        default = 0,
        metavar = 'N',
        type = int,
        dest = 'maxRoutes',
    ),
    ParseArgument('--checklist',
        help = 'Provide a checklist flow for the route.',
        action = 'store_true',
        default = False,
    ),
    ParseArgument('--x52-pro',
        help = 'Enable experimental X52 Pro MFD output.',
        action = 'store_true',
        default = False,
        dest = 'x52pro',
    ),
    ParseArgument('--prune-score',
        help = 'From the 3rd hop on, only consider routes which have at least this percentage of the current best route''s score.',
        dest = 'pruneScores',
        type = float,
        default = 0,
    ),
    ParseArgument('--prune-hops',
        help = 'Changes which hop --prune-score takes effect from.',
        default = 3,
        type = int,
        dest = 'pruneHops',
    ),
    ParseArgument('--progress', '-P',
        help = 'Show hop progress',
        default = False,
        action = 'store_true',
    ),
    ParseArgument('--supply',
        help = 'Only considers items which have at least this many units.',
        default = None,
        type = int,
    ),
    ParseArgument('--demand',
        help = 'Only considers items which have at least this much demand.',
        default = None,
        type = int
    ),
    ParseArgument('--summary',
        help = 'Summary layout of route instructions.',
        action = 'store_true',
    ),
    ParseArgument('--shorten',
        help = '(Requires --to) Find the shortest route with the best gpt.',
        action = 'store_true',
    ),
]

######################################################################
# Helpers


class Checklist:
    """
        Class for encapsulating display of a route as a series of
        steps to be 'checked off' as the user passes through them.
    """
    
    def __init__(self, tdb, cmdenv):
        self.tdb = tdb
        self.cmdenv = cmdenv
        self.mfd = cmdenv.mfd
    
    def doStep(self, action, detail = None, extra = None):
        self.stepNo += 1
        try:
            self.mfd.display(
                "#{} {}".format(self.stepNo, action),
                detail or "",
                extra or ""
            )
        except AttributeError:
            pass
        input(
            "   {:<3}: {}: "
            .format(
                self.stepNo,
                " ".join(item for item in (action, detail, extra) if item)
            )
        )
    
    def note(self, str, addBreak = True):
        print("(i) {} (i){}".format(str, "\n" if addBreak else ""))
    
    def run(self, route, cr):
        mfd = self.mfd
        stations, hops, jumps = route.route, route.hops, route.jumps
        lastHopIdx = len(stations) - 1
        gainCr = 0
        self.stepNo = 0
        
        heading = "(i) BEGINNING CHECKLIST FOR {} (i)".format(route.text(lambda x, y: y))
        print(heading, "\n", '-' * len(heading), "\n\n", sep = '')
        
        cmdenv = self.cmdenv
        if cmdenv.detail:
            print(route.summary())
            print()
        
        for idx in range(lastHopIdx):
            hopNo = idx + 1
            cur, nxt, hop = stations[idx], stations[idx + 1], hops[idx]
            sortedTradeOptions = sorted(
                hop[0],
                key=lambda tradeOption: tradeOption[1] * tradeOption[0].gainCr,
                reverse=True
            )
            
            # Tell them what they need to buy.
            if cmdenv.detail:
                self.note("HOP {} of {}".format(hopNo, lastHopIdx))
            
            self.note("Buy at {}".format(cur.name()))
            for (trade, qty) in sortedTradeOptions:
                self.doStep(
                        'Buy {:n} x'.format(qty),
                        trade.name(),
                        '@ {}cr / {} old'.format(
                            trade.costCr,
                            describeAge(trade.srcAge),
                ))
            if cmdenv.detail:
                self.doStep('Refuel')
            print()
            
            # If there is a next hop, describe how to get there.
            self.note(
                "Fly {}"
                .format(
                    " -> ".join(jump.name() for jump in jumps[idx])
                )
            )
            if idx < len(hops) and jumps[idx]:
                for jump in jumps[idx][1:]:
                    self.doStep('Jump to', jump.name())
            if cmdenv.detail:
                self.doStep('Dock at', nxt.text())
            print()
            
            self.note("Sell at {}".format(nxt.name()))
            for (trade, qty) in sortedTradeOptions:
                self.doStep(
                        'Sell {:n} x'.format(qty),
                        trade.name(),
                        '@ {:n}cr / {} old'.format(
                            trade.costCr + trade.gainCr,
                            describeAge(trade.dstAge),
                ))
            print()
            
            gainCr += hop[1]
            if cmdenv.detail and gainCr > 0:
                self.note("GAINED: {:n}cr, CREDITS: {:n}cr".format(
                            gainCr, cr + gainCr))
            
            if hopNo < lastHopIdx:
                print("\n--------------------------------------\n")
        
        if mfd:
            mfd.display('FINISHED',
                        "+{:n}cr".format(gainCr),
                        "={:n}cr".format(cr + gainCr))
            mfd.attention(3)
            from time import sleep
            sleep(1.5)


def expandForJumps(tdb, cmdenv, calc, origin, jumps, srcName, purpose):
    """
    Find all the stations you could reach if you made a given
    number of jumps away from the origin list.
    """

    assert jumps

    maxLyPer = cmdenv.emptyLyPer or cmdenv.maxLyPer
    avoidPlaces = cmdenv.avoidPlaces
    cmdenv.DEBUG0(
        "expanding {} reach from {} by {} jumps at {}ly per jump",
        srcName,
        origin.name(),
        jumps,
        maxLyPer,
    )

    # Preserve original behavior: --to uses stationsSelling, --from uses stationsBuying
    if srcName == "--to":
        tradingList = calc.stationsSelling
    elif srcName == "--from":
        tradingList = calc.stationsBuying
    else:
        raise Exception("Unknown src")

    # Ensure O(1) membership checks regardless of the underlying container type.
    trading_ids = tradingList if isinstance(tradingList, set) else set(tradingList)

    stations = set()
    origins, avoid = set((origin,)), set(place for place in avoidPlaces)

    for jump in range(jumps):
        if not origins:
            break
        if getattr(cmdenv, "debug", False):
            cmdenv.DEBUG1(
                "Ring {}: {}",
                jump,
                [sys.dbname for sys in origins]
            )
        thisJump, origins = origins, set()
        for system in thisJump:
            avoid.add(system)
            for stn in system.stations or ():
                if stn.ID not in trading_ids:
                    if getattr(cmdenv, "debug", False):
                        cmdenv.DEBUG2(
                            "X {}/{} not in trading list",
                            stn.system.dbname, stn.dbname,
                        )
                    continue
                if not checkStationSuitability(cmdenv, calc, stn):
                    if getattr(cmdenv, "debug", False):
                        cmdenv.DEBUG2(
                            "X {}/{} was not suitable",
                            stn.system.dbname, stn.dbname,
                        )
                    continue
                if getattr(cmdenv, "debug", False):
                    cmdenv.DEBUG2(
                        "- {}/{} meets requirements",
                        stn.system.dbname, stn.dbname,
                    )
                stations.add(stn)
            for dest, dist in tdb.genSystemsInRange(system, maxLyPer):
                if dest not in avoid:
                    origins.add(dest)

    if getattr(cmdenv, "debug", False):
        cmdenv.DEBUG0(
            "Expanded {} stations: {}",
            srcName,
            [stn.name() for stn in stations]
        )

    if not stations:
        if not cmdenv.emptyLyPer:
            extra = (
                "\nIf you are willing to make unladen jumps for the sake "
                "of a better route, consider using --empty."
            )
        else:
            extra = ""
        raise CommandLineError(
            "No {} stations with suitable trade data could be found "
            "within {} {}ly jump{} of {} that meet all of your critera.{}"
            .format(
                purpose, maxLyPer,
                jumps, "s" if jumps > 1 else "",
                origin.name(),
                extra,
            )
        )

    stations = list(stations)
    stations.sort(key=lambda stn: stn.ID)

    return stations

def checkForEmptyStationList(category, focusPlace, stationList, jumps):
    if stationList:
        return
    if jumps:
        raise NoDataError(
                "Local database has no price data for any "
                "stations within {} jumps of {} ({})".format(
                    jumps,
                    focusPlace.name(),
                    category,
        ))
    if isinstance(focusPlace, System):
        raise NoDataError(
                "Local database either has no price data for "
                "stations in {} ({}) or could not find any that "
                "met your requirements (e.g. pad-size). "
                "Check \"trade.py local -vv --ly 0 {}\"".format(
                    focusPlace.name(),
                    category,
                    focusPlace.name(),
        ))
    raise NoDataError(
            "Local database has no price data for {} ({})".format(
                focusPlace.name(),
                category,
    ))


def checkAnchorNotInVia(hops, anchorName, place, viaSet):
    """
    Ensure that '--to' or '--from' is not in the via set.
    """
    
    if hops != 2:
        return
    if isinstance(place, Station) and place in viaSet:
        raise CommandLineError(
            "{} used in {} and --via with only 2 hops".format(
                place.name(),
                anchorName,
        ))


def checkStationSuitability(cmdenv, calc, station, src = None):
    cmdenv.DEBUG2(
        "checking {} (ls={}, bm={}, pad={}, plt={}, flc={}, ody={}, mkt={}, shp={}) "
        "for {} suitability",
        station.name(),
        station.lsFromStar,
        station.blackMarket,
        station.maxPadSize,
        station.planetary,
        station.fleet,
        station.odyssey,
        station.market,
        station.shipyard,
        src or "any",
    )
    
    if station in cmdenv.avoidPlaces and src != "--from":
        if src:
            raise CommandLineError(
                "{} station {} is marked to avoid"
                .format(src, station.name())
            )
        return False
    if station.system in cmdenv.avoidPlaces and src != "--from":
        if src:
            raise CommandLineError(
                "{} station {} is in system listed in --avoid"
                .format(src, station.name())
            )
        return False
    if station.market == 'N':
        if src:
            raise CommandLineError(
                "{} station {} is flagged as having no market".format(
                    src, station.name()
                )
            )
        return False
    if not station.itemCount:
        if src:
            raise NoDataError(
                "No price data in local database "
                "for {} station: {}".format(
                    src, station.name(),
            ))
        return False
    if src != "--to" and station.ID not in calc.stationsSelling:
        if src:
            raise NoDataError(
                "No buying prices at {}."
                .format(station.name())
            )
        return False
    if src != "--from" and station.ID not in calc.stationsBuying:
        if src:
            raise NoDataError(
                "No selling prices at {}."
                .format(station.name())
            )
        return False
    mps = cmdenv.padSize
    if mps and not station.checkPadSize(mps):
        if src:
            raise CommandLineError(
                "{} station {} does not meet pad-size requirement.\n"
                "You specified: {}, Current data for station: {} ({})\n"
                "You can use \"trade.py station\" to correct this.".format(
                    src, station.name(),
                    mps, station.maxPadSize,
                    TradeDB.padSizesExt[station.maxPadSize],
            ))
        return False
    pla = cmdenv.planetary
    if pla and not station.checkPlanetary(pla):
        if src:
            raise CommandLineError(
                "{} station {} does not meet planetary requirement.\n"
                "You specified: {}, Current data for station: {} ({})\n"
                "You can use \"trade.py station\" to correct this.".format(
                    src, station.name(),
                    pla, station.planetary,
                    TradeDB.planetStatesExt[station.planetary],
            ))
        return False
    flc = cmdenv.fleet
    if flc and not station.checkFleet(flc):
        if src:
            raise CommandLineError(
                "{} station {} does not meet fleet carrier requirement.\n"
                "You specified: {}, Current data for station: {} ({})\n"
                "You can use \"trade.py station\" to correct this.".format(
                    src, station.name(),
                    flc, station.fleet,
                    TradeDB.fleetStatesExt[station.fleet],
            ))
        return False
    ody = cmdenv.odyssey
    if ody and not station.checkOdyssey(ody):
        if src:
            raise CommandLineError(
                "{} station {} does not meet odyssey requirement.\n"
                "You specified: {}, Current data for station: {} ({})\n"
                "You can use \"trade.py station\" to correct this.".format(
                    src, station.name(),
                    ody, station.odyssey,
                    TradeDB.odysseyStatesExt[station.odyssey],
            ))
        return False
    np = cmdenv.noPlanet
    if np and station.planetary != 'N':
        if src and src != "--from":
            raise CommandLineError(
                "{} station {} does not meet no-planet "
                "requirement.".format(
                    src, station.name(),
            ))
        return False
    bm = cmdenv.blackMarket
    if bm and station.blackMarket != 'Y':
        if src and src != "--from":
            raise CommandLineError(
                "{} station {} does not meet black-market "
                "requirement.".format(
                    src, station.name(),
            ))
        return False
    mls = cmdenv.maxLs
    if mls and (station.lsFromStar <= 0 or station.lsFromStar > mls):
        if src and src != "--from":
            raise CommandLineError(
                "{} station {} does not meet max-ls "
                "requirement.".format(
                    src, station.name(),
            ))
        return False
    maxAge, stnAge = cmdenv.maxAge, station.dataAge or float("inf")
    if maxAge and stnAge > maxAge:
        if src and src != "--from":
            raise CommandLineError(
                "{} station {} does not meet --age "
                "requirement.".format(
                    src, station.name(),
            ))
        return False
    return True


def filterStationSet(src, cmdenv, calc, stnList):
    if not stnList:
        return stnList
    if getattr(cmdenv, "debug", False):
        cmdenv.DEBUG0(
            "filtering {} station list: {}",
            src,
            ",".join(station.name() for station in stnList),
        )
    stnList = tuple(
        place for place in stnList
        if isinstance(place, System) or checkStationSuitability(cmdenv, calc, place, src)
    )
    if not stnList:
        # Preserve original message formatting (no behavior change)
        raise CommandLineError("No {src} station met your criteria.")
    return stnList

def checkOrigins(tdb, cmdenv, calc):
    # Compute eligibility once: stations must both sell and buy (same as suitability with src=None).
    eligible_ids = set(calc.stationsSelling) & set(calc.stationsBuying)

    if cmdenv.origPlace:
        if cmdenv.startJumps and cmdenv.startJumps > 0:
            cmdenv.origins = expandForJumps(
                tdb, cmdenv, calc,
                cmdenv.origPlace.system,
                cmdenv.startJumps,
                "--from", "starting",
            )
            cmdenv.origPlace = None
        elif isinstance(cmdenv.origPlace, System):
            cmdenv.DEBUG0("origPlace: System: {}", cmdenv.origPlace.name())
            if not cmdenv.origPlace.stations:
                raise CommandLineError(
                    "No stations at --from system, {}"
                    .format(cmdenv.origPlace.name())
                )
            cmdenv.origins = tuple(
                station
                for station in cmdenv.origPlace.stations
                if station.ID in eligible_ids and checkStationSuitability(cmdenv, calc, station)
            )
        else:
            cmdenv.DEBUG0("origPlace: Station: {}", cmdenv.origPlace.name())
            checkStationSuitability(cmdenv, calc, cmdenv.origPlace, '--from')
            cmdenv.origins = (cmdenv.origPlace,)
            cmdenv.startStation = cmdenv.origPlace
        checkForEmptyStationList(
            "--from", cmdenv.origPlace,
            cmdenv.origins, cmdenv.startJumps
        )
    else:
        if cmdenv.startJumps:
            raise CommandLineError("--start-jumps (-s) only works with --from")
        cmdenv.DEBUG0("using all suitable origins")
        cmdenv.origins = tuple(
            station
            for station in tdb.stationByID.values()
            if station.ID in eligible_ids and checkStationSuitability(cmdenv, calc, station)
        )

    if not cmdenv.startJumps and isinstance(cmdenv.origPlace, System):
        cmdenv.origins = filterStationSet(
            '--from', cmdenv, calc, cmdenv.origins
        )

    cmdenv.origSystems = tuple(set(
        stn.system for stn in cmdenv.origins
    ))

def checkDestinations(tdb, cmdenv, calc):
    cmdenv.destinations = None
    showProgress = bool(getattr(cmdenv, "progress", False))
    hb_interval = 0.5
    last_hb = 0.0
    spinner = ("|", "/", "-", "\\")
    spin_i = 0

    def heartbeat(seen, kept):
        nonlocal last_hb, spin_i
        if not showProgress:
            return
        now = time.time()
        if (now - last_hb) < hb_interval:
            return
        last_hb = now
        s = spinner[spin_i]
        spin_i = (spin_i + 1) % len(spinner)
        sys.stdout.write(
            f"\r{s} Scanning stations…  examined {seen:n}  kept {kept:n}"
        )
        sys.stdout.flush()

    if cmdenv.destPlace:
        if cmdenv.endJumps and cmdenv.endJumps > 0:
            cmdenv.destinations = expandForJumps(
                    tdb, cmdenv, calc,
                    cmdenv.destPlace.system,
                    cmdenv.endJumps,
                    "--to", "destination",
            )
            cmdenv.destPlace = None
        elif isinstance(cmdenv.destPlace, Station):
            cmdenv.DEBUG0("destPlace: Station: {}", cmdenv.destPlace.name())
            checkStationSuitability(cmdenv, calc, cmdenv.destPlace, '--to')
            cmdenv.destinations = (cmdenv.destPlace,)
        else:
            cmdenv.DEBUG0("destPlace: System: {}", cmdenv.destPlace.name())
            # Iterate with heartbeat instead of tuple-comprehension
            dests, seen, kept = [], 0, 0
            for station in cmdenv.destPlace.stations:
                seen += 1
                if checkStationSuitability(cmdenv, calc, station):
                    dests.append(station)
                    kept += 1
                heartbeat(seen, kept)
            cmdenv.destinations = tuple(dests)
            if showProgress:
                sys.stdout.write("\n"); sys.stdout.flush()
        checkForEmptyStationList(
                "--to", cmdenv.destPlace,
                cmdenv.destinations, cmdenv.endJumps
        )
    else:
        if cmdenv.endJumps:
            raise CommandLineError("--end-jumps (-e) only works with --to")
        cmdenv.DEBUG0("Using all available destinations")
        if cmdenv.goalSystem:
            dest = tdb.lookupPlace(cmdenv.goalSystem)
            cmdenv.goalSystem = dest.system

        if cmdenv.origPlace and cmdenv.maxJumpsPer == 0:
            stationSrc = chain.from_iterable(
                system.stations for system in cmdenv.origSystems
            )
        else:
            stationSrc = tdb.stationByID.values()

        # Pre-filter by eligibility to skip obviously ineligible stations
        eligible_ids = set(calc.stationsSelling) & set(calc.stationsBuying)

        # Iterate with heartbeat
        dests, seen, kept = [], 0, 0
        for station in stationSrc:
            seen += 1
            if station.ID in eligible_ids and checkStationSuitability(cmdenv, calc, station):
                dests.append(station)
                kept += 1
            heartbeat(seen, kept)
        cmdenv.destinations = tuple(dests)
        if showProgress:
            sys.stdout.write("\n"); sys.stdout.flush()

    if not cmdenv.endJumps and isinstance(cmdenv.destPlace, System):
        cmdenv.destinations = filterStationSet(
            '--to', cmdenv, calc, cmdenv.destinations
        )

    cmdenv.destSystems = tuple(set(
        stn.system for stn in cmdenv.destinations
    ))



def validateRunArguments(tdb, cmdenv, calc):
    """
        Process arguments to the 'run' option.
    """
    
    if cmdenv.credits < 0:
        raise CommandLineError("Invalid (negative) value for initial credits")
    # I'm going to allow 0 credits as a future way of saying "just fly"
    
    if cmdenv.routes < 1:
        raise CommandLineError(
            "Maximum routes has to be 1 or higher."
        )
    if cmdenv.routes > 1 and cmdenv.checklist:
        raise CommandLineError(
            "Checklist can only be applied to a single route."
        )
    
    if cmdenv.hops < 1:
        raise CommandLineError("Minimum of 1 hop required")
    if cmdenv.hops > 32:
        raise CommandLineError("Too many hops without more optimization")
    
    if cmdenv.maxJumpsPer < 0:
        raise CommandLineError("Negative jumps: you're already there?")
    if cmdenv.direct:
        cmdenv.hops = 1
        cmdenv.maxJumpsPer = cmdenv.maxLyPer = 10000
    
    if cmdenv.capacity is None:
        raise CommandLineError("Missing '--capacity'")
    if cmdenv.maxLyPer is None and not cmdenv.direct:
        raise CommandLineError("Missing '--ly-per'")
    if cmdenv.capacity < 0:
        raise CommandLineError("Invalid (negative) cargo capacity")
    if cmdenv.capacity > 1500:
        cmdenv.WARN("Capacity > 1500 not supported (you specified {})", cmdenv.capacity)
        cmdenv.WARN("Forcing jumps per hop to 1.")
        cmdenv.maxJumpsPer = 1
        if cmdenv.hops > 2:
            cmdenv.WARN("{} hops? Press [CTRL][C] to quit.", cmdenv.hops)
        if not cmdenv.supply:
            cmdenv.WARN("Please provide a '--supply' value.")
            cmdenv.supply = cmdenv.capacity * 10
            cmdenv.DEBUG0("'supply' minimum set to {}.", cmdenv.supply)
        if not cmdenv.demand:
            cmdenv.WARN("Please provide a '--demand' value.")
            cmdenv.demand = cmdenv.capacity * 10
            cmdenv.DEBUG0("'demand' minimum set to {}.", cmdenv.demand)
    #    raise CommandLineError(
    #        "Capacity > 1500 not supported (you specified {})"
    #        .format(cmdenv.capacity)
    #    )
    
    if cmdenv.limit and cmdenv.limit > cmdenv.capacity:
        raise CommandLineError("'limit' must be <= capacity")
    if cmdenv.limit and cmdenv.limit < 0:
        raise CommandLineError("'limit' can't be negative, silly")
    cmdenv.maxUnits = cmdenv.limit if cmdenv.limit else cmdenv.capacity
    
    if cmdenv.insurance:
        arbitraryInsuranceBuffer = 42
        if cmdenv.insurance >= (cmdenv.credits + arbitraryInsuranceBuffer):
            raise CommandLineError("Insurance leaves no margin for trade")
    
    if cmdenv.loop:
        if cmdenv.unique:
            raise CommandLineError("Cannot use --unique and --loop together")
        if cmdenv.direct:
            raise CommandLineError("Cannot use --direct and --loop together")
    
    if cmdenv.loopInt:
        if cmdenv.loopInt < 2:
            raise CommandLineError(
                "--loop-int must be 2 or higher to have any effect. "
            )
        if cmdenv.loopInt > cmdenv.hops and not cmdenv.unique:
            cmdenv.NOTE("--loop-int > hops implies --unique")
            cmdenv.unique = True
    
    if cmdenv.shorten:
        if cmdenv.loop:
            raise CommandLineError(
                "Cannot use --shorten and --loop together"
            )
        if not cmdenv.ending:
            raise CommandLineError(
                "--shorten only works with --to."
            )
    
    if cmdenv.goalSystem and not cmdenv.origPlace:
        raise CommandLineError("--towards requires --from")
    
    checkOrigins(tdb, cmdenv, calc)
    checkDestinations(tdb, cmdenv, calc)
    
    # If they're going --from and --to single systems, and they have
    # specified zero jumps then it's futile to try anything.
    if cmdenv.maxJumpsPer == 0 and not cmdenv.direct:
        if len(cmdenv.origSystems) == 1 and len(cmdenv.destSystems) == 1:
            if cmdenv.origSystems[0] != cmdenv.destSystems[0]:
                raise CommandLineError(
                    "Could not find any connections that didn't require at "
                    "least one jump and --jumps 0 specified."
                )
    
    origins, destns = cmdenv.origins or (), cmdenv.destinations or ()
    
    if cmdenv.hops == 1 and len(origins) == 1 and len(destns) == 1:
        if origins == destns:
            raise CommandLineError("Same to/from; more than one hop required.")
    
    avoidSet = set(cmdenv.avoidPlaces or ())
    viaSet = cmdenv.viaSet = set(cmdenv.viaPlaces)
    cmdenv.DEBUG0("Via: {}", viaSet)
    cmdenv.viaSet = filterStationSet('--via', cmdenv, calc, cmdenv.viaSet)
    checkAnchorNotInVia(cmdenv.hops, "--from", cmdenv.origPlace, viaSet)
    checkAnchorNotInVia(cmdenv.hops, "--to", cmdenv.destPlace, viaSet)
    
    viaSystems = set()
    for place in viaSet:
        if place in avoidSet or place.system in avoidSet:
            raise CommandLineError(
                '"--via {}" conflicts with --avoid'
                .format(place.name())
            )
        if isinstance(place, Station):
            viaSystems.add(place.system)
        else:
            viaSystems.add(place)
    
    if cmdenv.maxJumpsPer == 0 and viaSet and not cmdenv.direct:
        for via in viaSet:
            if via.system not in cmdenv.origSystems:
                raise CommandLineError(
                    "--via {} unreachable with --jumps 0"
                    .format(via.name())
                )
        cmdenv.origins = tuple(
            origin for origin in cmdenv.origins
            if origin.system in viaSystems
        )
        cmdenv.origSystems = tuple(
            origin.system for origin in cmdenv.origins
        )
        cmdenv.destinations = tuple(
            dest for dest in cmdenv.destinations
            if dest.system in viaSystems
        )
        cmdenv.destSystems = tuple(
            dest.system for dest in cmdenv.destinations
        )
    
    # How many of the hops do not have pre-determined stations. For example,
    # when the user uses "--from", they pre-determine the starting station.
    fixedRoutePoints = 0
    if cmdenv.origPlace:
        fixedRoutePoints += 1
    if cmdenv.destPlace:
        fixedRoutePoints += 1
    totalRoutePoints = cmdenv.hops + 1
    adhocRoutePoints = totalRoutePoints - fixedRoutePoints
    if len(viaSystems) > adhocRoutePoints:
        raise CommandLineError(
                "Route is not long enough for the list of '--via' "
                "destinations you gave. Reduce the vias or try again "
                "with '--hops {}' or greater.\n".format(
                    len(viaSet) + fixedRoutePoints - 1
                ))
    cmdenv.adhocHops = adhocRoutePoints - 1
    
    if cmdenv.unique and cmdenv.hops >= len(tdb.stationByID):
        raise CommandLineError(
            "Requested unique trip with more hops than there are stations..."
        )
    if cmdenv.unique:
        # if there's only one start and stop...
        if len(origins) == 1 and len(destns) == 1:
            if origins[0] is destns[0]:
                raise CommandLineError("Can't have same from/to with --unique")
        if viaSet:
            if len(origins) == 1 and origins[0] in viaSet:
                raise CommandLineError("Can't have --from station in --via list with --unique")
            if len(destns) == 1 and destns[0] in viaSet:
                raise CommandLineError("Can't have --to station in --via list with --unique")
    
    if cmdenv.mfd:
        cmdenv.mfd.display("Loading Trades")
    
    if cmdenv.pruneScores and cmdenv.pruneHops:
        if cmdenv.pruneScores > 99:
            raise CommandLineError("--prune-score value percentile exceed 99.")
        if cmdenv.pruneHops < 2:
            raise CommandLineError("--prune-hops must 2 or more.")
    else:
        cmdenv.pruneScores = cmdenv.pruneHops = 0

######################################################################


def filterByVia(routes, viaSet, viaStartPos):
    if not routes:
        return ()
    
    matchedRoutes = []
    partialRoutes = {}
    maxMet = 0
    for route in routes:
        met = 0
        for hop in route.route[viaStartPos:]:
            if hop in viaSet or hop.system in viaSet:
                met += 1
        if met > 0:
            if met >= len(viaSet):
                matchedRoutes.append(route)
            else:
                if met > maxMet:
                    partialRoutes[met] = []
                if met >= maxMet:
                    maxMet = met
                    partialRoutes[met].append(route)
    
    if matchedRoutes:
        return matchedRoutes, None
    
    if not maxMet:
        raise NoDataError(
                "No routes were found which matched your 'via' selections."
        )
    
    return partialRoutes[maxMet], (
            "SORRY: No runs visited all of your via destinations. "
            "Listing runs that matched at least {}.".format(
                    maxMet
            )
    )


def checkReachability(tdb, cmdenv):
    if cmdenv.direct:
        return
    srcSys, dstSys = cmdenv.origSystems, cmdenv.destSystems
    if len(srcSys) == 1 and len(dstSys) == 1:
        srcSys, dstSys = srcSys[0], dstSys[0]
        if srcSys != dstSys:
            maxLyPer = cmdenv.maxLyPer
            avoiding = tuple(
                avoid for avoid in cmdenv.avoidPlaces
                if isinstance(avoid, System)
            )
            route = tdb.getRoute(
                srcSys, dstSys, maxLyPer, avoiding,
            )
            if not route:
                raise CommandLineError(
                    "No route between {} and {} with a {}ly/jump limit."
                    .format(
                        srcSys.name(), dstSys.name(),
                        maxLyPer,
                    )
                )
            
            # Were there just not enough hops?
            jumpLimit = cmdenv.maxJumpsPer * cmdenv.hops
            routeJumps = len(route) - 1
            if jumpLimit < routeJumps:
                hopsRequired = math.ceil(routeJumps / cmdenv.maxJumpsPer)
                jumpsRequired = math.ceil(routeJumps / cmdenv.hops)
                raise CommandLineError(
                    "Shortest route between {src} and {dst} at {jumply} "
                    "ly per jump requires at least {minjumps} jumps. "
                    "Your current settings (--hops {hops} --jumps {jumps}) "
                    "allows a maximum of {jumplimit}.\n"
                    "\n"
                    "You may need --hops={althops} or --jumps={altjumps}.\n"
                    "\n"
                    "See also:\n"
                    " --towards (aka -T),"
                    " --start-jumps (-s),"
                    " --end-jumps (-e),"
                    " --direct.\n"
                    .format(
                        src = srcSys.name(),
                        dst = dstSys.name(),
                        jumply = cmdenv.maxLyPer,
                        minjumps = routeJumps,
                        hops = cmdenv.hops,
                        jumps = cmdenv.maxJumpsPer,
                        jumplimit = jumpLimit,
                        althops = hopsRequired,
                        altjumps = jumpsRequired,
                    )
                )


def routeFailedRestrictions(
        tdb, cmdenv, restrictTo, maxLs, hopNo
        ):
    """
    Generate exception text indicating we couldn't complete a
    route given the restrictions supplied. If the user has
    specified detail, check if there is a route at all.
    """
    
    places = list(
        set(
            chain.from_iterable(
                (place,) if isinstance(place, Station) else place.stations
                for place in restrictTo
            )
        )
    )
    places.sort(key = lambda stn: stn.dbname)
    
    dests = ", ".join(place.name() for place in places)
    
    return (
        "SORRY: Could not find any routes that delivered a profit to "
        "{} at hop #{}\n"
        "You may need to add more hops to your route or adjust your "
        "filters/restrictions.\n"
        .format(
            dests, hopNo + 1
        )
    )


def extraRouteProgress(routes):
    bestGain = max(routes, key = lambda route: route.gainCr).gainCr
    worstGain = min(routes, key = lambda route: route.gainCr).gainCr
    if bestGain != worstGain:
        gainText = "{:n}-{:n}cr gain".format(worstGain, bestGain)
    else:
        gainText = "{:n}cr gain".format(bestGain)
    
    bestGPT = int(max(routes, key = lambda route: route.gpt).gpt)
    worstGPT = int(min(routes, key = lambda route: route.gpt).gpt)
    if bestGPT != worstGPT:
        gptText = "{:n}-{:n}cr/ton".format(worstGPT, bestGPT)
    else:
        gptText = "{:n}cr/ton".format(bestGPT)
    
    return ".. {}, {}".format(gainText, gptText)

######################################################################
# Perform query and populate result set


def run(results, cmdenv, tdb):
    cmdenv.DEBUG1("loading trades")
    
    if tdb.tradingCount == 0:
        raise NoDataError("Database does not contain any profitable trades.")

    # Always show a friendly heads-up before heavy work begins.
    print("Searching for quality trades. This may take a few minutes. Please be patient.", flush=True)
    
    # Instantiate the calculator object
    calc = TradeCalc(tdb, cmdenv)
    
    validateRunArguments(tdb, cmdenv, calc)
    
    origPlace, viaSet = cmdenv.origPlace, cmdenv.viaSet
    stopStations = cmdenv.destinations
    goalSystem = cmdenv.goalSystem
    maxLs = cmdenv.maxLs
    
    # seed the route table with starting places
    startCr = cmdenv.credits - cmdenv.insurance
    routes = [
        Route(
            stations = (src,),
            hops = (),
            jumps = (),
            startCr = startCr,
            gainCr = 0,
            score = 0,
        )
        for src in cmdenv.origins
    ]
    
    numHops = cmdenv.hops
    lastHop = numHops - 1
    viaStartPos = 1 if origPlace else 0
    
    cmdenv.DEBUG1("numHops {}, vias {}, adhocHops {}",
                numHops, len(viaSet), cmdenv.adhocHops)
    
    results.summary = ResultRow()
    results.summary.exception = ""
    
    if cmdenv.loop:
        routePickPred = lambda route: \
            route.lastStation is route.firstStation
    elif cmdenv.shorten:
        if not cmdenv.destPlace:
            routePickPred = lambda route: \
                route.lastStation is route.firstStation
        elif isinstance(cmdenv.destPlace, System):
            routePickPred = lambda route: \
                route.lastSystem is cmdenv.destPlace
        else:
            routePickPred = lambda route: \
                route.lastStation is cmdenv.destPlace
    else:
        routePickPred = None
    
    pickedRoutes = []
    
    pruneMod = cmdenv.pruneScores / 100
    
    if cmdenv.loop:
        distancePruning = lambda rt, distLeft: \
            rt.lastSystem.distanceTo(rt.firstSystem) <= distLeft
    elif cmdenv.destPlace and not cmdenv.direct:
        distancePruning = lambda rt, distLeft: \
            any(
                stop for stop in stopSystems
                if rt.lastSystem.distanceTo(stop) <= distLeft
            )
    else:
        distancePruning = False
    
    if distancePruning:
        maxHopDistLy = cmdenv.maxJumpsPer * cmdenv.maxLyPer
        if not cmdenv.loop:
            stopSystems = {stop.system for stop in stopStations}
    
    for hopNo in range(numHops):
        restrictTo = None
        if hopNo == lastHop and stopStations:
            restrictTo = set(stopStations)
            manualRestriction = bool(cmdenv.destPlace)
        elif len(viaSet) > cmdenv.adhocHops:
            restrictTo = viaSet
            manualRestriction = True
        
        if distancePruning:
            preCrop = len(routes)
            distLeft = maxHopDistLy * (numHops - hopNo)
            routes = [rt for rt in routes if distancePruning(rt, distLeft)]
            if not routes:
                if pickedRoutes:
                    break
                raise NoDataError(
                    "No routes are in-range of any end stations at the end of hop {}"
                    .format(hopNo)
                )
            pruned = preCrop - len(routes)
            if pruned:
                cmdenv.NOTE("Pruned {} origins too far from any end stations", pruned)
        
        if hopNo >= 1 and (cmdenv.maxRoutes or pruneMod):
            routes.sort()
            if pruneMod and hopNo + 1 >= cmdenv.pruneHops and len(routes) > 10:
                crop = int(len(routes) * pruneMod)
                routes = routes[:-crop]
                cmdenv.NOTE("Pruned {} origins", crop)
            
            if cmdenv.maxRoutes and len(routes) > cmdenv.maxRoutes:
                routes = routes[:cmdenv.maxRoutes]
        
        if cmdenv.progress:
            extra = ""
            if hopNo > 0 and cmdenv.detail > 1:
                extra = extraRouteProgress(routes)
            print(
                "* Hop {:3n}: {:.>10n} origins {}"
                .format(hopNo + 1, len(routes), extra)
            )
        elif cmdenv.debug:
            cmdenv.DEBUG0("Hop {}...", hopNo + 1)
        
        try:
            newRoutes = calc.getBestHops(routes, restrictTo = restrictTo)
        except NoHopsError:
            if hopNo == 0 and len(cmdenv.origSystems) == 1:
                raise NoDataError(
                    "Couldn't find any trading links within {} x {}ly jumps of {}."
                    .format(
                        cmdenv.maxJumpsPer,
                        cmdenv.maxLyPer,
                        cmdenv.origSystems[0].name(),
                    )
                )
            raise NoDataError(
                "No routes had reachable trading links at hop #{}".format(hopNo + 1)
            )
        
        if not newRoutes:
            if pickedRoutes:
                break
            checkReachability(tdb, cmdenv)
            if hopNo > 0:
                if restrictTo and manualRestriction:
                    results.summary.exception += routeFailedRestrictions(
                        tdb, cmdenv, restrictTo, maxLs, hopNo
                    )
                    break
                results.summary.exception += (
                    "SORRY: Could not find profitable destinations "
                    "beyond hop #{:n}\n"
                    .format(hopNo + 1)
                )
                break
            if hopNo == 0:
                if cmdenv.origPlace and len(routes) == 1:
                    errText = (
                        "No profitable buyers found for the goods at {}.\n"
                        "\n"
                        "You may want to try:\n"
                        "  {} local \"{}\" --ly {} -vv --stations --trading"
                        .format(
                            routes[0].lastStation.name(),
                            sys.argv[0], cmdenv.origPlace.system.name(),
                            cmdenv.maxJumpsPer * cmdenv.maxLyPer,
                        )
                    )
                    if isinstance(cmdenv.origPlace, Station):
                        errText += (
                            "\n"
                            "or:\n"
                            "  {} market \"{}\" --sell -vv"
                            .format(
                                sys.argv[0], cmdenv.origPlace.name(),
                            )
                        )
                    raise NoDataError(errText)
        
        routes = newRoutes
        if routes and goalSystem:
            # Promote the winning route to the top of the list
            # while leaving the remainder of the list intact
            routes.sort(
                key = lambda route:
                    0 if route.lastSystem is goalSystem else 1
            )
            if routes[0].lastSystem is goalSystem:
                cmdenv.NOTE("Goal system reached!")
                routes = routes[:1]
                break
        
        if routePickPred:
            pickedRoutes.extend(
                route for route in routes if routePickPred(route)
            )
    
    if cmdenv.loop or cmdenv.shorten:
        cmdenv.DEBUG0("Using {} picked routes", len(pickedRoutes))
        routes = pickedRoutes
        # normalise the scores for fairness...
        for route in routes:
            cmdenv.DEBUG0(
                "{} hops, {} score, {} gpt",
                len(route.hops), route.score, route.gpt
            )
            route.score /= len(route.hops)
    
    if not routes:
        raise NoDataError(
            "No profitable trades matched your critera, "
            "or price data along the route is missing."
        )
    
    if viaSet:
        routes, caution = filterByVia(routes, viaSet, viaStartPos)
        if caution:
            results.summary.exception += caution + "\n"
    
    routes.sort()
    results.data = routes
    
    return results

######################################################################
# Transform result set into output


def render(results, cmdenv, tdb):
    exception = results.summary.exception
    if exception:
        print('#' * 76)
        print("\a{}".format(exception), end = "")
        print('#' * 76)
        print()
    
    routes = results.data
    
    for i in range(min(len(routes), cmdenv.routes)):
        print(routes[i].detail(cmdenv))
    
    # User wants to be guided through the route.
    if cmdenv.checklist:
        assert cmdenv.routes == 1
        cl = Checklist(tdb, cmdenv)
        cl.run(routes[0], cmdenv.credits)



# tradedangerous/db/adapter.py
from __future__ import annotations

from contextlib import contextmanager
from typing import Dict, Generator, Iterable, Optional, Tuple

from sqlalchemy import select, func
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Session

# Local engine + ORM (authoritative)
from .engine import make_engine_from_config, get_session_factory  # uses env/CWD-resolved db_config.ini by default
from .orm_models import System, Station, Item, StationItem  # canonical models
from .paths import resolve_db_config_path

# ---- Public factory ---------------------------------------------------------

def get_adapter_if_enabled(cfg_path: Optional[str] = None) -> "TradeDBReadAdapter | None":
    """
    Return an adapter when [database] backend != 'sqlite', else None.
    - No engine/session created at import: construction is lazy.
    - This is called by tradedb.py (thin gate).
    """
    import configparser, os
    if cfg_path is None:
        cfg_path = str(resolve_db_config_path())
    cfg = configparser.ConfigParser()
    if not os.path.exists(cfg_path):
        return None
    with open(cfg_path, "r", encoding="utf-8") as fh:
        cfg.read_file(fh)
    backend = (cfg.get("database", "backend", fallback="sqlite") or "sqlite").strip().lower()
    if backend == "sqlite":
        return None

    # Engine is created lazily via the property below to honour "no side-effects at import".
    return TradeDBReadAdapter(cfg_path)

# ---- Adapter (read-only) ----------------------------------------------------

class TradeDBReadAdapter:
    """
    Very small, read-only façade over SQLAlchemy for legacy TradeDB reads:
      - systems() list
      - lookup system by name (case-insensitive)
      - station by (system_id, station_name) (case-insensitive)
      - average selling/buying prices (used by trade_cmd at detail>1)
    """
    def __init__(self, cfg_path: str):
        self._cfg_path = cfg_path
        self._engine: Optional[Engine] = None
        self._Session = None  # sessionmaker

    # Lazy engine/session factory (no import-time work)
    @property
    def Session(self):
        if self._Session is None:
            engine = make_engine_from_config(self._cfg_path)
            self._engine = engine
            self._Session = get_session_factory(engine)
        return self._Session

    @contextmanager
    def session(self) -> Generator[Session, None, None]:
        Session = self.Session
        with Session() as s:
            yield s

    # ---- Reads mapped to ORM ------------------------------------------------

    def list_system_rows(self) -> Iterable[Tuple[int, str, float, float, float, Optional[int]]]:
        """
        Shape matches legacy _loadSystems SELECT:
        (system_id, name, pos_x, pos_y, pos_z, added_id)
        """
        with self.session() as s:
            rows = s.execute(
                select(
                    System.system_id,
                    System.name,
                    System.pos_x,
                    System.pos_y,
                    System.pos_z,
                    System.added_id,
                )
            )
            for r in rows:
                yield (r.system_id, r.name, r.pos_x, r.pos_y, r.pos_z, r.added_id)

    def system_by_name(self, name_ci: str) -> Optional[Tuple[int, str, float, float, float, Optional[int]]]:
        """
        Case-insensitive name match for System.
        """
        with self.session() as s:
            row = s.execute(
                select(
                    System.system_id, System.name, System.pos_x, System.pos_y, System.pos_z, System.added_id
                ).where(func.upper(System.name) == func.upper(func.trim(func.cast(name_ci, System.name.type))))
            ).first()
            if not row:
                return None
            return (row.system_id, row.name, row.pos_x, row.pos_y, row.pos_z, row.added_id)

    def station_by_system_and_name(
        self, system_id: int, station_name_ci: str
    ) -> Optional[Tuple[int, int, str, int, str, str, str, str, str, str, str, str, str, int]]:
        """
        Return the single Station row by system + name (CI).
        Shape matches legacy _loadStations row consumed by Station(...):
          (station_id, system_id, name,
           ls_from_star, market, blackmarket, shipyard,
           max_pad_size, outfitting, rearm, refuel, repair, planetary, type_id)
        """
        with self.session() as s:
            r = s.execute(
                select(
                    Station.station_id,
                    Station.system_id,
                    Station.name,
                    Station.ls_from_star,
                    Station.market,
                    Station.blackmarket,
                    Station.shipyard,
                    Station.max_pad_size,
                    Station.outfitting,
                    Station.rearm,
                    Station.refuel,
                    Station.repair,
                    Station.planetary,
                    Station.type_id,
                ).where(
                    Station.system_id == system_id,
                    func.upper(Station.name) == func.upper(func.trim(func.cast(station_name_ci, Station.name.type))),
                )
            ).first()
            if not r:
                return None
            return (
                r.station_id,
                r.system_id,
                r.name,
                r.ls_from_star,
                r.market,
                r.blackmarket,
                r.shipyard,
                r.max_pad_size,
                r.outfitting,
                r.rearm,
                r.refuel,
                r.repair,
                r.planetary,
                r.type_id,
            )

    def average_selling(self) -> Dict[int, int]:
        """
        {item_id: avg_supply_price>0}
        Mirrors the legacy SQL used in TradeDB.getAverageSelling.
        """
        with self.session() as s:
            rows = s.execute(
                select(
                    Item.item_id,
                    func.IFNULL(func.avg(StationItem.supply_price), 0),
                )
                .select_from(Item.__table__.outerjoin(
                    StationItem, (Item.item_id == StationItem.item_id) & (StationItem.supply_price > 0)
                ))
                .where(StationItem.supply_price > 0)
                .group_by(Item.item_id)
            )
            return {int(item_id): int(avg_cr) for (item_id, avg_cr) in rows}

    def average_buying(self) -> Dict[int, int]:
        """
        {item_id: avg_demand_price>0}
        Mirrors the legacy SQL used in TradeDB.getAverageBuying.
        """
        with self.session() as s:
            rows = s.execute(
                select(
                    Item.item_id,
                    func.IFNULL(func.avg(StationItem.demand_price), 0),
                )
                .select_from(Item.__table__.outerjoin(
                    StationItem, (Item.item_id == StationItem.item_id) & (StationItem.demand_price > 0)
                ))
                .where(StationItem.demand_price > 0)
                .group_by(Item.item_id)
            )
            return {int(item_id): int(avg_cr) for (item_id, avg_cr) in rows}



from pathlib import Path

from .. import fs
from ..commands.exceptions import CommandLineError
from . import PluginException, ImportPluginBase

class ImportPlugin(ImportPluginBase):
    """
    Plugin that imports multiple .price files at once
    """
    
    PRICE_GLOB = "*.prices"
    BATCH_FILE = "batch.prices"
    
    pluginOptions = {
        'files': 'Path to some .price files to import (file=file1;file2)',
        'folder': 'Path to a folder containing .price files (folder=pathToFolder)'
    }
    
    def __init__(self, tdb, tdenv):
        super().__init__(tdb, tdenv)
    
    """
    Returns the newest file.
    """
    def file_get_newer(self, a, b):
        if a.stat().st_mtime > b.stat().st_mtime:
            return a
        else:
            return b
    
    """
    Returns a list of stations in a file
    """
    def file_get_stations(self, a):
        contents = a.read_text()
        stations = []
        for line in contents.split("\n"):
            if "@" in line:
                stations.append(line[2:])
        return stations
    
    """
    Given a list of files, does analysis to determine
    whether they are duplicates.
    """
    def sanitize_files(self, files):
        stations_seen = {}
        
        for f in files:
            filePath = Path(f)
            if not filePath.is_file():
                raise CommandLineError("File not found: {}".format(
                    str(filePath)
                ))
            
            stations = self.file_get_stations(filePath)
            
            # Does not support resolving multiple stations (yet)
            if len(stations) > 1:
                raise PluginException("EDMC Batch unable to process files with multiple stations. Use normal import.")
            
            for s in stations:
                if s in stations_seen:
                    cur_file = stations_seen[s]
                    # Set the newer file as the one we'll use.
                    stations_seen[s] = self.file_get_newer(cur_file, f)
                else:
                    stations_seen[s] = f
        
        return stations_seen.values()
    
    """
    Set the environment for the import command to import our files
    """
    def set_environment(self, files):
        tdenv = self.tdenv
        path_list = files
        fs.ensurefolder(tdenv.tmpDir)
        batchfile = tdenv.tmpDir / Path(self.BATCH_FILE)
        if batchfile.exists():
            batchfile.unlink()
        # We now have a list of paths. Add all contents to a new file
        temp_file = open(batchfile, "w")
        
        for f in path_list:
            contents = f.read_text()
            temp_file.writelines("# File: {}\n".format(f))
            temp_file.write(contents)
        
        # Set the file we're reading from to the temp file
        tdenv.filename = str(batchfile.absolute())
    
    def split_files(self, files):
        file_list = self.getOption("files").split(";")
        path_list = []
        for f in file_list:
            path_list.append(Path(f))
        return path_list
    
    def get_prices_in_folder(self, directory):
        folderToSplit = Path(directory)
        if not folderToSplit.is_dir():
            raise CommandLineError("Path is not a directory: {}".format(
                str(folderToSplit)
            ))
        
        filesInFolder = folderToSplit.glob(self.PRICE_GLOB)
        file_list = []
        for i in filesInFolder:
            file_list.append(i)
        
        return file_list
    
    def run(self):
        # Grab files option and validate
        files = self.getOption("files")
        folder = self.getOption("folder")
        
        if files is None and folder is None:
            raise PluginException("Argument Required")
        
        # The list of files to import...
        file_list = []
        
        if files:
            file_list.extend(self.split_files(files))
        if folder:
            file_list.extend(self.get_prices_in_folder(folder))
        
        # Split into a path list, verify all paths are good.
        self.set_environment(self.sanitize_files(file_list))
        return True




# tradedangerous/commands/update_cmd.py — DEPRECATED (no-op)
# This command no longer edits station price lists directly.
#
# How to maintain price/station data going forward:
#   • Import authoritative data via:   trade import -P spansh  |  trade import -P eddblink
#   • Solo/offline players (manual capture via EDMC plugin):
#       TradeDangerous DB-Update for EDMC → https://github.com/bgol/UpdateTD
#
# This module accepts and ignores any arguments/switches and exits immediately.

from __future__ import annotations
from .parsing import ParseArgument

# ---- Command metadata ----
help = "DEPRECATED: no longer used. See deprecation banner when run."
name = "update"
epilog = None

# Swallow unknown switches/args (loader must honor this flag).
acceptUnknown = True

# No DB access is needed for this no-op command.
wantsTradeDB = False
usesTradeData = False

# Accept ANY number of positional args and ignore them (prevents parser errors).
arguments = (
    ParseArgument(
        "args",
        help="(deprecated) ignored",
        nargs="*",
        type=str,
    ),
)
# No switches; unknown switches will be ignored via acceptUnknown.
switches = tuple()


def _banner() -> str:
    return (
        "\n"
        "=== DEPRECATION NOTICE: update =====================================\n"
        "This command no longer edits station price lists and does not modify the DB.\n"
        "• Import data via:               trade import -P eddblink | -P spansh\n"
        "• Solo/offline capture via EDMC: TradeDangerous DB-Update → https://github.com/bgol/UpdateTD\n"
        "=====================================================================\n"
    )


def run(results, cmdenv, tdb=None):
    """
    No-op implementation: print banner and exit immediately.
    All arguments/switches are ignored by design.
    """
    banner = _banner()
    try:
        cmdenv.NOTE("{}", banner)
    except Exception:
        print(banner)
    return None


def render(results, cmdenv, tdb=None):
    # No output beyond the banner emitted in run().
    return



# tradedangerous/db/lifecycle.py
from __future__ import annotations

from pathlib import Path
from typing import Iterable, List, Tuple, Optional, Dict

from sqlalchemy import inspect, text
from sqlalchemy.engine import Engine
from sqlalchemy.schema import MetaData


# --------------------------------------------------------------------
# Utilities
# --------------------------------------------------------------------

def is_sqlite(engine: Engine) -> bool:
    """Return True if the SQLAlchemy engine is using SQLite dialect."""
    return engine.dialect.name == "sqlite"


def _user_tables(engine: Engine) -> Iterable[str]:
    """List user (non-internal) tables present in the current database."""
    insp = inspect(engine)
    names = insp.get_table_names()
    if is_sqlite(engine):
        names = [n for n in names if not n.startswith("sqlite_")]
    return names


def is_empty(engine: Engine) -> bool:
    """True when no user tables exist (via SQLAlchemy Inspector)."""
    return len(list(_user_tables(engine))) == 0


# --------------------------------------------------------------------
# (Re)creation helpers — prefer explicit paths; discovery is fallback
# --------------------------------------------------------------------

def _read_sql_file(sql_path: Path) -> str:
    """Read the provided SQL file (authoritative SQLite schema)."""
    if not sql_path.exists():
        raise FileNotFoundError(f"SQLite schema file not found: {sql_path}")
    return sql_path.read_text(encoding="utf-8")


def _read_legacy_sql() -> str:
    """Fallback: locate the legacy SQLite schema SQL if an explicit path was not provided."""
    candidates = [
        Path(__file__).resolve().parents[1] / "templates" / "TradeDangerous.sql",
        Path.cwd() / "tradedangerous" / "templates" / "TradeDangerous.sql",
        Path.cwd() / "TradeDangerous.sql",
    ]
    for p in candidates:
        if p.exists():
            return p.read_text(encoding="utf-8")
    raise FileNotFoundError("TradeDangerous.sql not found in expected locations.")


def _execute_sql_script(engine: Engine, script: str) -> None:
    """Execute a multi-statement SQL script using sqlite3's executescript()."""
    with engine.begin() as conn:
        raw_conn = conn.connection  # DB-API connection (sqlite3.Connection)
        raw_conn.executescript(script)


def _create_sqlite_from_legacy(engine: Engine, sql_path: Optional[Path] = None) -> None:
    """Create the SQLite schema by executing the legacy SQL (explicit path preferred)."""
    if sql_path is not None:
        sql = _read_sql_file(sql_path)
    else:
        sql = _read_legacy_sql()
    _execute_sql_script(engine, sql)


# --------------------------------------------------------------------
# Public resets
# --------------------------------------------------------------------

def reset_sqlite(engine: Engine, db_path: Path, sql_path: Optional[Path] = None) -> None:
    """
    Reset the SQLite schema by rotating the DB file and recreating from legacy SQL.

    Steps:
      1) Dispose the SQLAlchemy engine to release pooled sqlite file handles.
      2) Rotate the on-disk database file to a .old sibling (idempotent; cross-device safe).
      3) Ensure the target directory exists.
      4) Recreate the schema using the provided canonical SQL file (or fallback discovery).

    Notes:
      - Rotation naming preserves your historic convention:
            TradeDangerous.db  →  TradeDangerous.old
      - If no DB file exists, rotation is a no-op.
    """
    # 1) Release any open file handles held by the connection pool
    try:
        engine.dispose()
    except Exception:
        pass  # best-effort

    # 2) Rotate DB → .old (idempotent, cross-device safe)
    db_path = db_path.resolve()
    old_path = db_path.with_suffix(".old")
    try:
        if db_path.exists():
            try:
                if old_path.exists():
                    old_path.unlink()
            except Exception:
                # If removal of old backup fails, continue and let rename/copy raise if necessary
                pass

            try:
                db_path.rename(old_path)
            except OSError:
                # Cross-device or locked: copy then unlink
                import shutil
                shutil.copy2(db_path, old_path)
                try:
                    db_path.unlink()
                except Exception:
                    # If unlink fails, leave both; schema recreate will still run on db_path
                    pass
    except Exception:
        # Rotation shouldn't prevent schema recreation; continue
        pass

    # 3) Make sure parent directory exists
    try:
        db_path.parent.mkdir(parents=True, exist_ok=True)
    except Exception:
        pass

    # 4) Recreate schema from canonical SQL
    _create_sqlite_from_legacy(engine, sql_path=sql_path)

def reset_mariadb(engine: Engine, metadata: MetaData) -> None:
    """
    Drop all tables and recreate using ORM metadata (MariaDB/MySQL),
    with FOREIGN_KEY_CHECKS disabled during the operation.

    This avoids FK-ordering issues and makes resets deterministic.
    """
    # Use a transactional connection for the whole reset
    with engine.begin() as conn:
        # Disable FK checks for the duration of drop/create
        conn.execute(text("SET FOREIGN_KEY_CHECKS=0"))
        try:
            metadata.drop_all(bind=conn)
            metadata.create_all(bind=conn)
        finally:
            # Always restore FK checks
            conn.execute(text("SET FOREIGN_KEY_CHECKS=1"))



# --------------------------------------------------------------------
# Unified reset entry point (dialect hidden from callers)
# --------------------------------------------------------------------

def reset_db(engine: Engine, *, db_path: Path, sql_path: Optional[Path] = None) -> str:
    """
    Reset the database schema for the given engine in a dialect-appropriate way.

    Caller MUST pass the canonical on-disk `db_path` for SQLite and SHOULD pass `sql_path`.
    (No path deduction is attempted here beyond optional SQL discovery.)

    Returns a short action string for logs/tests.
    """
    dialect = engine.dialect.name.lower()

    if dialect == "sqlite":
        reset_sqlite(engine, db_path=db_path, sql_path=sql_path)
        return "sqlite:rotated+recreated"

    if dialect in ("mysql", "mariadb"):
        # Resolve ORM metadata internally to avoid dialect branching at call sites.
        from tradedangerous.db import orm_models
        reset_mariadb(engine, orm_models.Base.metadata)
        return f"{dialect}:reset"

    raise RuntimeError(f"Unsupported database backend: {engine.dialect.name}")


# --------------------------------------------------------------------
# Sanity checks (seconds-only, no deep I/O)
# --------------------------------------------------------------------

# NOTE: 'Added' removed from core set (being deprecated)
_CORE_TABLES: Tuple[str, ...] = (
    "System",
    "Station",
    "Category",
    "Item",
    "StationItem",
)

def _core_tables_and_pks_ok(engine: Engine) -> Tuple[bool, List[str]]:
    """
    T1 + T2: Required core tables exist and have primary keys.
    Returns (ok, problems).
    """
    problems: List[str] = []
    insp = inspect(engine)

    existing = set(insp.get_table_names())
    missing = [t for t in _CORE_TABLES if t not in existing]
    if missing:
        problems.append(f"missing tables: {', '.join(missing)}")
        return False, problems

    for t in _CORE_TABLES:
        pk = insp.get_pk_constraint(t) or {}
        cols = pk.get("constrained_columns") or []
        if not cols:
            problems.append(f"missing primary key on {t}")

    return (len(problems) == 0), problems


def _seed_counts_ok(engine: Engine) -> Tuple[bool, List[str]]:
    """
    T4: Minimal seed/anchor rows must exist.
      - Category > 0
      - System   > 0
    """
    problems: List[str] = []
    with engine.connect() as conn:
        for tbl in ("Category", "System"):
            cnt = conn.execute(text(f"SELECT COUNT(*) FROM {tbl}")).scalar() or 0
            if cnt <= 0:
                problems.append(f"{tbl} is empty")

    return (len(problems) == 0), problems


def _connectivity_ok(engine: Engine) -> bool:
    """T0: Cheap connectivity probe (redundant if T4 runs, but negligible)."""
    try:
        with engine.connect() as conn:
            conn.execute(text("SELECT 1"))
        return True
    except Exception:
        return False


# --------------------------------------------------------------------
# Orchestration (policy) — may call buildCache
# --------------------------------------------------------------------

def ensure_fresh_db(
    backend: str,
    engine: Engine,
    data_dir: Path,
    metadata: MetaData | None,
    mode: str = "auto",
    *,
    tdb=None,
    tdenv=None,
    rebuild: bool = True,
) -> Dict[str, str]:
    """
    Ensure a *sane, populated* database exists (seconds-only checks).

      Checks:
        - T0: connectivity
        - T1/T2: core tables exist and have PKs
        - T4: seed rows exist in Category and System

    Actions:
      - mode == "force"                      → rebuild via buildCache(...) (if rebuild=True)
      - mode == "auto" and not sane         → rebuild via buildCache(...) (if rebuild=True)
      - not sane and rebuild == False       → action = "needs_rebuild" (NEVER rebuild)
      - sane and mode != "force"            → action = "kept"

    Returns a summary dict including:
      - backend, mode, action, sane (Y/N), and optional reason.

    NOTE:
      - When a rebuild is required but rebuild=True and (tdb/tdenv) are missing,
        a ValueError is raised (preserves current semantics).
      - When rebuild=False, the function NEVER calls buildCache and never raises
        for missing tdb/tdenv. It simply reports the status.
    """
    summary: Dict[str, str] = {
        "backend": (backend or engine.dialect.name).lower(),
        "mode": mode,
        "action": "kept",
        "sane": "Y",
    }

    # T0: cheap connectivity
    if not _connectivity_ok(engine):
        summary["reason"] = "connectivity-failed"
        summary["sane"] = "N"
        if mode == "auto":
            mode = "force"

    # T1+T2: structure; T4: seeds
    if summary["sane"] == "Y":
        structure_ok, struct_problems = _core_tables_and_pks_ok(engine)
        if not structure_ok:
            summary["sane"] = "N"
            summary["reason"] = "; ".join(struct_problems) or "structure-invalid"
        else:
            seeds_ok, seed_problems = _seed_counts_ok(engine)
            if not seeds_ok:
                summary["sane"] = "N"
                reason = "; ".join(seed_problems) or "seeds-missing"
                summary["reason"] = f"{summary.get('reason','')}; {reason}".strip("; ").strip()

    sane = (summary["sane"] == "Y")
    must_rebuild = (mode == "force") or (not sane)

    # If nothing to do, return immediately.
    if not must_rebuild:
        summary["action"] = "kept"
        return summary

    # Caller explicitly requested no rebuild: report and exit.
    if not rebuild:
        summary["action"] = "needs_rebuild"
        return summary

    # From here on, behavior matches the original: rebuild via buildCache.
    if tdb is None or tdenv is None:
        raise ValueError("ensure_fresh_db needs `tdb` and `tdenv` to rebuild via buildCache")
        
    from tradedangerous.cache import buildCache

    buildCache(tdb, tdenv)
    summary["action"] = "rebuilt"
    return summary



import io
import os
import pytest
import re

from tradedangerous.cli import trade
from .helpers import copy_fixtures, regex_findin, replace_stdin

PROG = "trade"


def setup_module():
    copy_fixtures()


class TestTradeRun:
    def test_run1(self, capsys):
        # Disable colour output for deterministic matching
        os.environ["TD_NO_COLOR"] = "1"
        trade([
            PROG, "run",
            "--capacity=10", "--credits=10000",
            "--from=sol/abr", "--jumps-per=3",
            "--ly-per=10.5", "--no-planet"
        ])
        captured = capsys.readouterr()
        assert "Sol/Abraham Lincoln: 10 x Hydrogen Fuel," in captured.out
        assert "Sol/Burnell Station: 2 x Silver," in captured.out
        assert "560cr (213/ton)" in captured.out

    @pytest.mark.slow
    def test_run2(self, capsys):
        os.environ["TD_NO_COLOR"] = "1"
        trade([
            PROG, "run", "-vv", "--progress", "--empty=82",
            "--cap=212", "--jumps=4", "--cr=2153796", "--from=sol/abr",
            "--hops=6", "--ls-m=8000", "--sup=10000",
            "--pad=L", "--ly=25", "--prune-hop=3", "--prune-sc=40"
        ])
        captured = capsys.readouterr()
        assert regex_findin(r"=> est [\d\s,]+cr total", captured.out)

    @pytest.mark.slow
    def test_run3(self, capsys):
        """Testing --checklist"""
        os.environ["TD_NO_COLOR"] = "1"
        STEPS = 37
        with replace_stdin(io.StringIO('\n' * STEPS)):
            trade([
                PROG, "run", "-vv", "--progress", "--empty=82", "--checklist",
                "--cap=212", "--jumps=4", "--cr=2153796", "--from=sol/abr",
                "--hops=6", "--ls-m=8000", "--sup=10000",
                "--pad=L", "--ly=25", "--prune-hop=3", "--prune-sc=40"
            ])
        captured = capsys.readouterr()
        assert "BEGINNING CHECKLIST FOR Sol/Abraham Lincoln -> LHS 449/Fisher Point" in captured.out
        assert "35 : Sell 212 x Polymers" in captured.out

    def test_run4(self, capsys):
        os.environ["TD_NO_COLOR"] = "1"
        trade([
            PROG, "run",
            "--capacity=10", "--credits=10000",
            "--from=sol/abr", "--jumps-per=3",
            "--ly-per=10.5", "--start-jumps=2"
        ])
        captured = capsys.readouterr()
        # Legacy route verification (plain text, no ANSI)
        assert "Sol/Haberlandt Survey -> Sol/Durrance Camp" in captured.out
        assert "  Sol/Haberlandt Survey: 5 x Reactive Armour," in captured.out
        assert "  Sol/Ehrlich City: 10 x Building Fabricators," in captured.out
        assert re.search(r"Sol/Durrance Camp \+10,?925cr \(728/ton\)", captured.out)


from __future__ import annotations
from .exceptions import CommandLineError
from .parsing import ParseArgument
from ..cache import buildCache
from ..tradedb import TradeDB

######################################################################
# Parser config

help = 'Build TradeDangerous cache file from sources'
name = 'buildcache'
epilog = (
        'TD will normally do this for you automatically whenever '
        'it detects changes to one or more source file; most end-'
        'users will never need to use this command.\n'
        'N.B. This process is destructive: '
        'any data in the .db that is not reflected in the '
        'source files will be lost.'
)
wantsTradeDB = False  # Cause we're about to frak with it.
arguments = [
]
switches = [
    ParseArgument(
        '--sql', default = None, dest = 'sqlFilename',
        help = 'Specify SQL script to execute.',
    ),
    ParseArgument(
        '--prices', default = None, dest = 'pricesFilename',
        help = 'Specify the prices file to load.',
    ),
    ParseArgument(
        '--force', '-f', default = False, action = 'store_true',
        dest = 'force',
        help = 'Overwrite existing file',
    ),
    ParseArgument(
        '--ignore-unknown', '-i',
        default = False, action = 'store_true',
        dest = 'ignoreUnknown',
        help = (
            "Data for systems, stations and items that are not "
            "recognized is reported as warning but skipped."
        ),
    ),
]

######################################################################
# Helpers

######################################################################
# Perform query and populate result set


def run(results, cmdenv, tdb: TradeDB):
    """
    BRUTE-FORCE rebuild of the cache/database.

    Semantics preserved:
      - If DB exists and --force not given => error
      - SQL file must exist
      - Performs a full destructive rebuild

    Implementation change:
      - Delegates to tradedangerous.db.lifecycle.ensure_fresh_db with mode='force'
        so all backend-specific checks and rebuild steps run via the central path.
    """
    # Deprecation note: keep short and visible but non-fatal.
    print("NOTE: 'buildcache' is deprecated. Prefer 'update' or importer plugins. "
          "Proceeding with a forced rebuild via db.lifecycle.ensure_fresh_db().")

    # Honor legacy safety: require --force to overwrite an existing DB file.
    if not cmdenv.force and tdb.dbPath.exists():
        raise CommandLineError(
            f"SQLite3 database '{tdb.dbFilename}' already exists.\n"
            "Either remove the file first or use the '-f/--force' option."
        )

    # Ensure the SQL source exists (buildCache ultimately relies on this path).
    if not tdb.sqlPath.exists():
        raise CommandLineError(f"SQL File does not exist: {tdb.sqlFilename}")

    # Force a rebuild through the lifecycle helper (works for both backends).
    from tradedangerous.db.lifecycle import ensure_fresh_db

    ensure_fresh_db(
        backend=tdb.engine.dialect.name if getattr(tdb, "engine", None) else "sqlite",
        engine=getattr(tdb, "engine", None),
        data_dir=tdb.dataPath,
        metadata=None,
        mode="force",
        tdb=tdb,
        tdenv=cmdenv,
        rebuild=True,
    )

    return None




# tradedangerous/db/locks.py
# -----------------------------------------------------------------------------
# Advisory lock helpers (MariaDB/MySQL) — per-station serialization
#
# SQLite compatibility:
#   - On SQLite (or any unsupported dialect), all helpers become NO-OPs and
#     behave as if the lock was immediately acquired (yield True). This lets
#     shared code run unchanged across backends.
#
# Usage (both writers must use the SAME key format):
#   from tradedangerous.db.locks import station_advisory_lock
#
#   with sa_session_local(session_factory) as s:
#       # (optional) set isolation once per process elsewhere:
#       # s.execute(text("SET SESSION TRANSACTION ISOLATION LEVEL READ COMMITTED")); s.commit()
#       with station_advisory_lock(s, station_id, timeout_seconds=0.2, max_retries=4) as got:
#           if not got:
#               # processor: defer/requeue work for this station and continue
#               return
#           with s.begin():
#               # do per-station writes here...
#               pass
# -----------------------------------------------------------------------------

from __future__ import annotations

import time
from contextlib import contextmanager
from typing import Iterator

from sqlalchemy import text
from sqlalchemy.orm import Session

__all__ = [
    "station_advisory_lock",
    "acquire_station_lock",
    "release_station_lock",
    "station_lock_key",
]

# Precompiled SQL (MySQL/MariaDB only)
_SQL_GET_LOCK     = text("SELECT GET_LOCK(:k, :t)")
_SQL_RELEASE_LOCK = text("SELECT RELEASE_LOCK(:k)")

def _is_lock_supported(session: Session) -> bool:
    """
    Return True if the current SQLAlchemy session is bound to a backend that
    supports advisory locks via GET_LOCK/RELEASE_LOCK (MySQL/MariaDB).
    """
    try:
        name = (session.get_bind().dialect.name or "").lower()
    except Exception:
        name = ""
    return name in ("mysql", "mariadb")
    
def _ensure_read_committed(session: Session) -> None:
    """
    Ensure the session is using READ COMMITTED for subsequent transactions.
    - Applies only to MySQL/MariaDB.
    - No-ops on SQLite/others.
    - Only sets it if NOT already inside a transaction (affects next txn).
    """
    if not _is_lock_supported(session):
        return
    try:
        # Only set if we're not already in a transaction; otherwise it would
        # affect the next transaction, not the current one.
        if not session.in_transaction():
            session.execute(text("SET SESSION TRANSACTION ISOLATION LEVEL READ COMMITTED"))
            # No explicit commit needed; this is a session-level setting.
    except Exception:
        # Best-effort; if this fails we just proceed with the default isolation.
        pass

def station_lock_key(station_id: int) -> str:
    """
    Return the advisory lock key used by both writers for the same station.
    Keep this format identical in all writers (processor + Spansh).
    """
    return f"td.station.{int(station_id)}"

def acquire_station_lock(session: Session, station_id: int, timeout_seconds: float) -> bool:
    """
    Try to acquire the advisory lock for a station on THIS DB connection.

    Returns:
        True  -> acquired within timeout (or NO-OP True on unsupported dialects)
        False -> timed out (lock held elsewhere)

    Notes:
        - Advisory locks are per-connection. Use the same Session for acquire,
          the critical section, and release.
        - On SQLite/unsupported dialects, this is a NO-OP that returns True.
    """
    if not _is_lock_supported(session):
        return True  # NO-OP on SQLite/unsupported backends

    key = station_lock_key(station_id)
    row = session.execute(_SQL_GET_LOCK, {"k": key, "t": float(timeout_seconds)}).first()
    # MariaDB/MySQL GET_LOCK returns 1 (acquired), 0 (timeout), or NULL (error)
    return bool(row and row[0] == 1)

def release_station_lock(session: Session, station_id: int) -> None:
    """
    Release the advisory lock for a station on THIS DB connection.
    Safe to call in finally; releasing a non-held lock is harmless.

    On SQLite/unsupported dialects, this is a NO-OP.
    """
    if not _is_lock_supported(session):
        return  # NO-OP on SQLite/unsupported backends

    key = station_lock_key(station_id)
    try:
        session.execute(_SQL_RELEASE_LOCK, {"k": key})
    except Exception:
        # Intentionally swallow — RELEASE_LOCK may return 0/NULL if not held.
        pass

@contextmanager
def station_advisory_lock(
    session: Session,
    station_id: int,
    timeout_seconds: float = 0.2,
    max_retries: int = 4,
    backoff_start_seconds: float = 0.05,
) -> Iterator[bool]:
    """
    Context manager to acquire/retry/release a per-station advisory lock.

    Resilience improvement:
      - If no transaction is active on the Session, this helper will OPEN ONE,
        so the lock is taken on the same physical connection the ensuing DML uses.
        In that case, it will COMMIT on normal exit, or ROLLBACK if an exception
        bubbles out of the context block.
      - If a transaction is already active, this helper does NOT touch txn
        boundaries; caller remains responsible for commit/rollback.

    Yields:
        acquired (bool): True if acquired within retry policy;
                         True immediately on unsupported dialects (NO-OP);
                         False if not acquired on supported backends.
    """
    # Fast-path NO-OP for SQLite/unsupported dialects
    if not _is_lock_supported(session):
        try:
            yield True
        finally:
            pass
        return

    # If we can still influence the next txn, prefer READ COMMITTED for shorter waits.
    _ensure_read_committed(session)

    # Pin a connection if caller hasn't already begun a transaction.
    started_txn = False
    txn_ctx = None
    if not session.in_transaction():
        txn_ctx = session.begin()
        started_txn = True

    got = False
    try:
        # Attempt with bounded retries + exponential backoff.
        attempt = 0
        while attempt < max_retries:
            if acquire_station_lock(session, station_id, timeout_seconds):
                got = True
                break
            time.sleep(backoff_start_seconds * (2 ** attempt))
            attempt += 1

        # Hand control to caller
        yield got

        # If we created the transaction and no exception occurred, commit it.
        if started_txn and got:
            try:
                session.commit()
            except Exception:
                # If commit fails, make sure to roll back so we don't leak an open txn.
                session.rollback()
                raise
    except Exception:
        # If we created the transaction and an exception escaped the block, roll it back.
        if started_txn and session.in_transaction():
            try:
                session.rollback()
            except Exception:
                # Swallow secondary rollback failures; original exception should propagate.
                pass
        raise
    finally:
        # Always release the advisory lock if we acquired it.
        if got:
            try:
                release_station_lock(session, station_id)
            except Exception:
                # Lock releases are best-effort; don't mask user exceptions.
                pass

        # If we opened a txn context object (older SA versions), ensure it's closed.
        # (Harmless if already committed/rolled back above.)
        if started_txn and txn_ctx is not None:
            try:
                txn_ctx.close()
            except Exception:
                pass



import pytest

from tradedangerous.cli import trade
from tradedangerous.commands.exceptions import UsageError
from .helpers import copy_fixtures, regex_findin, remove_fixtures

PROG = "trade"


def setup_module():
    copy_fixtures()


def teardown_module():
    remove_fixtures()


class TestTrade:
    def test_local_help(self):
        with pytest.raises(UsageError):
            trade([PROG, "local", "-h"])
    
    def test_local_sol(self, capsys):
        trade([PROG, "local", "--ly=10", "--detail", "sol"])
        captured = capsys.readouterr()
        assert "Sol       0" in captured.out
        assert "Ehrlich City" in captured.out
    
    def test_sell(self, capsys):
        trade([PROG, "sell", "--near=sol", "hydrogen fuel"])
        captured = capsys.readouterr()
        assert "Sol/Mars High" in captured.out
    
    def test_buy(self, capsys):
        trade([PROG, "buy", "--near=sol", "hydrogen fuel"])
        captured = capsys.readouterr()
        assert "Cost      Units DistLy Age/days" in captured.out
    
    def test_export_station(self, capsys):
        trade([PROG, "export", "-T", "System"])
        captured = capsys.readouterr()
        assert "NOTE: Export Table 'System'" in captured.out
        # TODO: check that System.csv has a fresh date

    @pytest.mark.skip(reason="station command deprecated; now prints deprecation banner")
    def test_station_remove(self, capsys):
        pytest.skip("station command deprecated; no longer edits DB directly")

    @pytest.mark.skip(reason="station command deprecated; now prints deprecation banner")
    def test_station_add(self, capsys):
        pytest.skip("station command deprecated; no longer edits DB directly")
    
    def test_nav(self, capsys):
        trade([PROG, "nav", "--ly-per=50", "sol", "Shinrarta Dezhra"])
        captured = capsys.readouterr()
        assert "System            JumpLy" in captured.out
        assert "Shinrarta Dezhra   47" in captured.out
    
    def test_market(self, capsys):
        trade([PROG, "market", "sol/abr"])
        captured = capsys.readouterr()
        assert regex_findin("Item[ ]{3,}Buying Selling[ ]{2,}Supply", captured.out)
        assert "Hydrogen Fuel" in captured.out
        assert regex_findin("Water[ ]{3,}323", captured.out)
    
    @pytest.mark.slow
    def test_import_edcd(self, capsys):
        trade([PROG, "import", "-P=edcd", "--opt=commodity"])
        captured = capsys.readouterr()
        assert regex_findin(r"NOTE: Found \d+ item\(s\)", captured.out)
